import sys
from pymongo import DESCENDING, ASCENDING
from bson.objectid import ObjectId
from .mongo_types import *
from pprint import pprint
from pymongoose.logger import Logger
import math

schemas = {}
debug_log = True
database = None

def set_schemas(db, schemas_re, _debug_log_=True):
    """
    ## Description:
    Function to be called were you control your models registration
    # Parameters
    ------------
    - db: Database
    MongoDB connection database
    - schemas_re: dict
    A dictionary with all your registerd schemas.
    f.e: {"users": User.schema}
    - _debug_log_: bool
    A variable to set logs on or off
    Default True
    """
    global database
    global schemas
    global debug_log
    database = db
    schemas = schemas_re
    debug_log = _debug_log_
    if debug_log:
        Logger.set_terminal_color("blue")
        print("MongoDB Schemas: ")
        pprint(schemas)
        Logger.set_terminal_color("reset")

def get_schema_by_name(name:str) -> dict:
    if name in schemas:
        return schemas[name]
    else: return None

def exists(schema, query):
    """
	Check if document exists in collection
	# Parameters
	------------
	### schema: Mongo collection name
		Mongo collection name where query will be executed
    ### query: dict
        Dictionary containing query
    # Returns
    ------------
    - boolean
	"""
    return database[schema].count_documents(query, limit = 1) > 0

def _get_clean_schema(schema, pop):
    isList = False
    aux_schema = schema.copy()
    if pop.find(".") > 0:
        populate_list = pop.replace(" ", "").split(".")
        for pop1 in populate_list:
            if type(aux_schema) is list: 
                isList = True
                aux_schema = aux_schema[0]

            aux_schema = aux_schema[pop1]
    else:
        aux_schema = aux_schema[pop]

    if type(aux_schema) is list:
        isList = True
        aux_schema = aux_schema[0]

    return aux_schema, isList

def _populate(schema, populate, aggregate, parent=""):
    wasList = False
    for pop in populate: #Iterate in each populate object or str
        if type(pop) is not dict: #If is a str is a simple lookup and unwind
            aux_schema, isList = _get_clean_schema(schema, pop)
            e = {
                "$lookup": {
                    "from": aux_schema["ref"],
                    "localField": parent + pop,
                    "foreignField": "_id",
                    "as": parent + pop
                }
            }
            aggregate.append(e)

            if not isList:
                unwind = {
                    "path": f"${parent + pop}",
                    "preserveNullAndEmptyArrays": True
                }
                aggregate.append({"$unwind": unwind})
            
        else: 
            aux_schema, isList = _get_clean_schema(schema, pop["path"])
            
            if isList:  #If is list first unwind path dessired
                aggregate.append(
                    {
                        "$unwind": {
                            "path": f"${parent}{pop['path'] if not pop['path'].find('.') else pop['path'].split('.')[0]}",
                            "preserveNullAndEmptyArrays": True
                        }
                    }
                )
                wasList = True
            ##Add respective pipeline with select and match
            pipeline = [
                {
                    "$match": {
                        "$expr":{
                            "$eq": ["$_id", "$$var"]
                        }
                    }
                },
            ]
            if "select" in pop:
                project = {}
                for s in pop["select"]:
                    project[s]= 1
                
                pipeline.append({"$project": project})

            if "match" in pop:
                pipeline.append({"$match": pop["match"]})
            

            e = {
                "$lookup": {
                    "from": aux_schema["ref"],
                    "let": {f"var": f"${parent + pop['path']}"},
                    "pipeline": pipeline,
                    "as": parent + pop["path"]
                }
            }

            aggregate.append(e)
            #Unwind result
            unwind = {
                "path": f"${parent + pop['path']}",
                "preserveNullAndEmptyArrays": True
            }
            aggregate.append({"$unwind": unwind})

            wL = False

            if "options" in pop: #If has attached populates
                
                ##Recursively call to function with respectie schema 

                wL = _populate(schemas[aux_schema["ref"]], pop["options"], aggregate, parent + pop["path"] + ".") 

            ## If current isList is true
            ## If len(parent) > 0 then current parent should be added in _id else _id = "$_id._id"
                ## path: {"$push": "$parent.path"}
                ## doc: {"$first": "$$ROOT"}
                ## add $replaceRoot with $_id, parent: {"$mergeObjects": ["$doc", {"_id": "$_id"}, {"$parent": {"$mergeObjects": ["$parent", {"path": "$path"}]}}]}

            if isList: #Group and replace to have a correct object
                group = {}
                _id = {}

                parentAux = parent
                if len(parent) > 0:
                    parentAux = parent[:-1]
                    _id = {
                        "_id": "$_id",
                        parentAux: f"${parentAux}._id"
                    }
                else: 
                    _id = "$_id._id" if wL else "$_id"

                popAux = pop["path"]
                if popAux.find(".") > 0:
                    popAux = popAux.split(".")[0]

                group["_id"] = _id
                group[popAux] = {"$push": f"${parent}{popAux}"}
                group["doc"] = {"$first": "$$ROOT"}

                aggregate.append({
                    "$group": group
                })

                replaceRoot = {}
                if len(parent) > 0:
                    replaceRoot = {
                        "$mergeObjects": [
                            "$doc",
                            {"_id": "$_id"},
                            {
                                parentAux: {
                                    "$mergeObjects": [f"$doc.{parentAux}", {popAux: f"${popAux}"}]
                                }
                            }
                        ]
                    }
                else:
                    replaceRoot = {
                        "$mergeObjects": [
                            "$doc",
                            {"_id": "$_id"},
                            
                            {popAux: f"${popAux}"}
                            
                        ]
                    }
                
                aggregate.append({
                    "$replaceRoot": {"newRoot": replaceRoot}
                })
                

    return wasList

def _convert_id_to_object_id(id) -> ObjectId:
    if type(id) is not ObjectId:
        id = ObjectId(id)
    return id

def count_documents(schema: str, query: dict) -> int:
    global schemas
    retval = -1
    try:
        schema_name = schema
        retval = database[schema_name].count_documents(query)
    except:
        raise MongoException(message="Error counting document(s)", mongoError=MongoError.Bad_action, bt=sys.exc_info())

    return retval

def find(schema: str, query: dict, select = {}, populate=None, one=False, skip = 0, limit=None, sort=None):
    global schemas
    schema_name = schema
    """
	Find a document inside a collection
	# Parameters
	------------
    ### schema: str
        Schema Name created in order to manage mongo documents
    ### query: dict
        Dictionary containing query
    ### select: dict
        Dictionary containing requested fields
        defaults to: {}
    ### populate: dict
        Dictionary containing lookup structure of the fields required to populate
        defaults to: None
    ### one: bool
        Variable to select if return one ore many
        defaults to: False
    ### skip: int
        Integer to skip to 'n' values to the left in the collection
        defaults to: 0
    ### limit: int
        Integer to limit number of documents returned from the collection
        defaults to: infinite
    ### sort: dict
        Dictionary to set an order of documents based on a field
        defaults to: None
    # Returns
    ------------
    - cursor if one is False else dict
	"""
    try:
        if "_id" in query and type(query["_id"]) is not dict:
            query["_id"] = _convert_id_to_object_id(query["_id"])
        
        
        sort_key, sort_value = "_id", ASCENDING
        if sort is not None:
                    sort_key = list(sort.keys())[0]
                    sort_value = ASCENDING if sort[sort_key] == 1 else DESCENDING

        schema = schemas[schema]
        retval = {}
        if populate is None:
            if one:
                retval = database[schema_name].find_one(query, select)
            else:
                if sort is not None:
                    if limit is None:
                        retval = database[schema_name].find(query, select).skip(skip).sort(sort_key, sort_value)
                    else:
                        retval = database[schema_name].find(query, select).skip(skip).limit(limit).sort(sort_key, sort_value)
                else:
                    if limit is None:
                        retval = database[schema_name].find(query, select).skip(skip)
                    else:
                        retval = database[schema_name].find(query, select).skip(skip).limit(limit)
        else:
            aggregate = [
                {"$match" : query}
            ]
        
            _populate(schema, populate, aggregate, "")
            
            if select is not None:
                aggregate.append({
                    "$project": select
                })

            if(skip > 0):
                aggregate.append({
                    "$skip": skip
                })
            
            if(limit is not None):
                aggregate.append({
                    "$limit": limit
                })

            if(sort is not None):
                aggregate.append({
                    "$sort": {
                        f"{sort_key}": sort_value
                    }
                })
            

            retval = database[schema_name].aggregate(aggregate)

            
        if populate is not None and one:
            retval = list(retval)
            if len(retval) > 0:
                return retval[0]
            else:
                return None
        else:
            return retval
        
        
    except:
        raise MongoException(message="Error finding document(s)", mongoError=MongoError.Bad_action, bt=sys.exc_info())

def find_by_id(schema, id, select = {}, populate=None):
    """
	Find a document inside a collection by _id
    (Same as find({_id:id}))
	# Parameters
	------------
    ### schema: dict
        Schema created in order to manage mongo documents
    ### id: str
        ObjectId of the document
    ### select: dict
        Dictionary containing requested fields
        defaults to: {}
    ### populate: dict
        Dictionary containing lookup structure of the fields required to populate
        defaults to: None
    # Returns
    ------------
    - dict
	"""
    if type(id) is not dict:
        id = _convert_id_to_object_id(id)
    
    retval = find(schema, {"_id": id}, select, populate, one=True)
    return retval

def insert_one(schema, object: dict) -> ObjectId:
    try:
        retval = database[schema].insert_one(object)

        return retval.inserted_id
    except:
        raise MongoException(message=f"Error saving object in {schema}", mongoError=MongoError.Bad_action)

def aggregate(schema, aggregate):
    """
	Find a document inside a collection
	# Parameters
	------------
	### collection: Mongo collection
		Mongo collection where query will be executed
    ### aggregate: list
        Aggregation pipeline that is required
    # Returns
    ------------
    - coursor if one is False else dict
	"""
    return database[schema].aggregate(aggregate)

def update(schema, query, update, many = False):
    """
	Update a document inside collection
	# Parameters
	------------
    ### schema: dict
        Schema created in order to manage mongo documents
    ### query: dict
        Dictionary containing query
    ### update: dict
        Dictionary containing mongodb update structure
    ### many: bool
        Variable to select if update one ore many
        defaults to: False
    # Returns
    ------------
    - Number of documents updated
	"""
    try:
        if "_id" in query:
            query["_id"] = _convert_id_to_object_id(query["_id"])

        if not many:
            retval = database[schema].update_one(query, update)
            return retval.modified_count
        else:
            retval = database[schema].update_many(query, update)
            return retval.modified_count
    except:
        raise MongoException(sys.exc_info()[0],  message="Error updating document(s)", mongoError=MongoError.Bad_actio)

def delete(schema, query, many = False):
    """
	Delete documents inside collection
	# Parameters
	------------
	### schema: dict
        Schema created in order to manage mongo documents
    ### query: dict
        Dictionary containing query
    ### many: bool
        Variable to select if delete one ore many
        defaults to: False
    # Returns
    ------------
    - number of documents deleted
	"""
    try:
        if "_id" in query:
            query["_id"] = _convert_id_to_object_id(query["_id"])

        if many:
            retval = database[schema].delete_many(query)
            return retval.deleted_count
        else:
            database[schema].delete_one(query)
            return 1
    except:
        raise MongoException(message="Error deleting document(s)", mongoError=MongoError.Bad_action, bt=sys.exc_info()[0])

def get_cursor_length(cursor):
    return len(list(cursor))


colors = {
	"red":
	   "\x1B[31m",
	"yellow":
	   "\x1B[33m",
	"green":
	   "\x1B[32m",
	"blue":
	   "\x1B[34m",
	"reset":
	   "\x1B[0m",
	None:
	   "\x1B[0m",
}

class Logger():
	@staticmethod
	def printWarn(text: str):
		print(f"{colors['yellow']}[PYMONGOOSE WARN]: {text}{colors['reset']}")

	@staticmethod
	def printLog(text: str):
		print(f"{colors['blue']}[PYMONGOOSE LOG]: {text}{colors['reset']}")

	@staticmethod
	def printError(text: str):
		print(f"{colors['red']}[PYMONGOOSE ERROR]: {text}{colors['reset']}")

	@staticmethod
	def printSuccess(text: str):
		print(f"{colors['green']}[PYMONGOOSE SUCCESS]: {text}{colors['reset']}")

	@staticmethod
	def set_terminal_color(color: str):
		print(colors[color])


import sys, traceback
from enum import Enum
import datetime
import bson.json_util as json_util
from pymongoose.logger import Logger
from pymongoose import methods
from bson import ObjectId
import json


class Types(Enum):
	"""
	Enum to be used on your schemas creation
	"""
	String      = 1
	Number      = 2
	Date        = 3
	ObjectId    = 4
	Boolean     = 5

class MongoError(Enum):
	"""
	Mongo error Enum used in mongo exception
	"""
	Empty_dataset	= 1
	Required_field 	= 2
	Bad_type 		= 3
	Bad_action		= 4

class MongoException(Exception):
	"""
	Mongo exception class to be used in mongo methods
	"""
	message = ""
	mongoError = None
	bt = ""

	def __init__(self, *args, **kwargs):
		self.message = kwargs["message"]
		self.mongoError = kwargs["mongoError"]
		exc_type, exc_obj, exc_tb = sys.exc_info()
		super().__init__(
			# f"Exception ocurred in line = {exc_tb.tb_lineno}\n",
			f"Offense ocurred = {MongoError(self.mongoError).name}\n" +
			f"Thrown message = {self.message}"
		)
		traceback.print_exc()
		if methods.debug_log:
			Logger.printError(
				# f"Exception ocurred in line = {exc_tb.tb_lineno}\n",
				f"Offense ocurred = {MongoError(self.mongoError).name}\n" +
				f"Thrown message = {self.message}"
			)	

class Schema(object):
	"""
	Parent class of any schema that is created
	"""
	id = None
	schema = {}
	schema_name = None

	def __init__(self, schema_name, schema, initial_values = {}):
		"""
		### schema_name: str
			Name set to the schema (must be plural)
		### schema: dict
			Schema generated with types, defaults and/or required fields
		### initial_values: dict
			Dictionary containing initial values of this schema, must contain same names as the schema has
			defaults to: {}
		"""
		self.schema = schema
		self.schema_name = schema_name
		id = initial_values["id"] if "id" in initial_values else ObjectId()
		self.id = ObjectId(id) if type(id) is str else id

		is_empty = False
		if "empty" in initial_values:
			is_empty = initial_values["empty"]

		
		if not is_empty:
			if len(initial_values) == 0:
				initial_values = {}
			
			clean_schema = {}
			for key in self.schema.keys():
				clean_schema[key] = self.get_default_value(key, initial_values)

			self.__dict__.update(clean_schema)
	
	def fromJson(self, json_obj):
		for key in self.schema.keys():
			if methods.debug_log:
				Logger.printLog(f"Setting {key}")
			setattr(self, key, self.get_default_value(key, json_obj))
			
		if methods.debug_log:
			Logger.printLog(f"Setting _id")
		self.id = self.get_default_value("_id", json_obj)

	def toJson (self, full=True):
		"""
		Method to convert current Schema object to a dictionary
		# Parameters
		------------
		### full: bool
			If true, will attach id to dictionary
			defaults to: True
		# Returns
		------------
		- dict
		"""
		json_obj = {}

		for key in self.schema.keys():
			json_obj[key] = getattr(self, key)

		if full:
			json_obj["id"] = self.id

		return json_obj

	def save(self, id = None):
		"""
		Function to save current object into a mongoDB collection
		# Parameters
		------------
		### id: ObjectId
			Set a forced id to save method
			defaults to: None
		# Returns
		------------
		- ObjectId -> Saved id
		"""
		if not self.validate_required(self.toJson(False), self.schema):
			raise MongoException(message="Required fields missing", mongoError=MongoError.Required_field)
		elif not self.validate_type(self.toJson(False), self.schema):
			raise MongoException(message="Type fields are wrong", mongoError=MongoError.Bad_type, bt=sys.exc_info())

		else:
			json_obj = self.toJson(False)
			if id is not None:
				json_obj["_id"] = id

			retval = methods.insert_one(self.schema_name, json_obj)
			self.id = retval
		return self.id

	def parse_schema_value(self, value):
		"""
		From current schema, will get default value if exists, None if not.
		# Parameters:
		--------------
		- value: string
		Name of field
		"""
		isList = False
		valAux = value
		if type(value) is list:
			isList = True
			valAux = value[0]
		
		if "default" in valAux:
			return valAux["default"] if not isList else [valAux["default"]]
		else:
			dictionary = {}
			for k in valAux:
				if type(valAux[k]) is dict:
					e = self.parse_schema_value(valAux[k])
					dictionary[k] = e
			if isList:
				return [] if len(dictionary) == 0 else [dictionary]
			else:
				return None if len(dictionary) == 0 else dictionary

	def extract(self, key, json_obj):
		"""
		Extract a value from a dictionary, if it doesn't exists then will get default value, else None
		# Parameters:
		------------
		- key: String
		Required key from dictionary
		- json_obj: dict
		Dictionary where it will be looked
		# Returns: [Value] or None
		"""
		return json_obj[key] if key in json_obj else self.get_default_value(key, [])

	def type_validator(self, type, value) -> bool:
		if type == Types.String:
			return True if type(value) is str else False
		elif type == Types.Number:
			return True if type(value) is float or type(value) is int else False
		elif type == Types.Date:
			return True if type(value) is datetime.datetime else False
		else: return True

	def _item_type_check(self, key, type, item_type):
		if item_type == None or item_type == dict:
			if methods.debug_log:
				Logger.printLog(f"key {key} doesn't come in json, returning True")
			return True

		if type == Types.Number:
			if item_type is int or item_type is float:
				return True
			else:
				if methods.debug_log:
					Logger.printError(f"key {key} has an incorrect type") 
				return False
		elif type == Types.String:
			if item_type is str:
				return True
			else:
				if methods.debug_log:
					Logger.printError(f"key {key} has an incorrect type") 
				return False
		elif type == Types.Date:
			if "datetime.datetime" in str(item_type):
				return True
			else:
				if methods.debug_log:
					Logger.printError(f"key {key} has an incorrect type") 
				return False
		elif type == Types.ObjectId:
			if item_type is ObjectId:
				return True
			else:
				if methods.debug_log:
					Logger.printError(f"key {key} has an incorrect type") 
				return False
		elif type == Types.Boolean:
			if item_type is bool:
				return True
			else:
				if methods.debug_log:
					Logger.printError(f"key {key} has an incorrect type")
				return False

	def _validate_type_cycle(self, scAux, json_obj, last_key):

		if "type" in scAux:
			tp = type(json_obj) if json_obj is not None else None
			retval = self._item_type_check(last_key, scAux["type"], tp)

			if methods.debug_log:
				if retval:
					Logger.printSuccess(f"(type) key={last_key} in schema has a correct value type")
				else:
					Logger.printError(f"(type) key={last_key} in schema has an incorrect value type")
			return retval

		retval = True
		for k in scAux:
			retval = True
			if type(scAux[k]) is dict or type(scAux[k]) is list:
				if "type" in scAux[k]:
					tp = type(json_obj[k]) if k in json_obj and json_obj[k] is not None else None
					retval = self._item_type_check(k, scAux[k]["type"], tp)
					
				elif type(scAux[k]) is list or dict in list(map(type, scAux[k].values())):
					retval = self.validate_type(
						json_obj[k] if k in json_obj and json_obj[k] is not None else
							[{}] if type(scAux[k]) is list else {},
						scAux[k], 
						k
					)
				else:
					if methods.debug_log:
						Logger.printWarn(f"(type) key={k} in schema doesn't contain any type, returning True")
					retval = True

				if not retval:
					if methods.debug_log:
						Logger.printError(f"(type) key={k} in schema has an incorrect value type")
					return retval
				
				if methods.debug_log:
						Logger.printSuccess(f"(type) key={k} in schema has a correct value type")
			else:
				if methods.debug_log:
					Logger.printWarn(f"(type) key={k} schema is not a dict, returning True")
			
		return retval

	def validate_type(self, json_obj: dict, sc=schema, last_key = None) -> bool:
		scAux = sc
		is_list = False
		if type(sc) is list:
			scAux = sc[0]
			is_list = True

		retval = False
		if is_list:
			if len(json_obj) == 0:
				retval = self._validate_type_cycle(scAux, {}, last_key)
			else:
				for i, obj in enumerate(json_obj):
					retval = self._validate_type_cycle(scAux, obj, last_key)

					if retval:
						if methods.debug_log:
							Logger.printSuccess(f"(type) key '{last_key}-{i}' has a valid type")
					else:
						return False

		else:
			retval = self._validate_type_cycle(scAux, json_obj, last_key)

		return retval

	def _validate_required_cycle(self, scAux, json_obj) -> bool:
		if "required" in scAux and scAux["required"]:
			return json_obj is not None

		for k in scAux:
			retval = True
			if type(scAux[k]) is dict or type(scAux[k]) is list:
				if "required" in scAux[k] and scAux[k]["required"]:
					if not k in json_obj or json_obj[k] is None:
						if methods.debug_log:
							Logger.printError(f"(req) key '{k}' is required but value doesn't come")
						return False
					else:
						retval = True
				elif type(scAux[k]) is list or dict in list(map(type, scAux[k].values())):
						retval = self.validate_required(
							json_obj[k] if k in json_obj and json_obj[k] is not None else
								[{}] if type(scAux[k]) is list else {}, 
							scAux[k], 
							k
						)
				else:
					if methods.debug_log:
						Logger.printLog(f"(req) key={k} doesn't contain required value, setting by default in False")
					retval = True

				if retval:
					if methods.debug_log:
						Logger.printSuccess(f"(req) key '{k}' has a valid argument")

				if not retval: 
					if methods.debug_log:
						Logger.printError(f"(req) key '{k}' is required but value doesn't come")
					return retval
			else:
				if methods.debug_log:
					Logger.printLog(f"(req) key={k} current schema isn't a dict, setting by default in False")

		return retval

	def validate_required(self, json_obj: dict, sc=schema, last_key="") -> bool:
		"""
		Check if current schema object contains all required fields dicted by schema
		# Parameters:
		------------
		- json_obj: dict
		Dictionary where it will be looked
		# Returns: bool
		"""
		scAux = sc
		is_list = False
		if type(sc) is list:
			scAux = sc[0]
			is_list = True

		retval = False
		if is_list:
			if len(json_obj) == 0:
				retval = self._validate_required_cycle(scAux, {})
			else:
				for i, obj in enumerate(json_obj):
					retval = self._validate_required_cycle(scAux, obj)

					if retval:
						if methods.debug_log:
							Logger.printSuccess(f"(req) key '{last_key}-{i}' has a valid argument")
					else:
						return False

		else:
			retval = self._validate_required_cycle(scAux, json_obj)

		return retval
		
	def get_default_value(self, name, kwargs):
		"""
		Wrapper function for parse_schema_value
		# Parameters:
		----------
		- name: str
		Name of the variable looking for
		- kwargs: tuple
		Array of variables that may contain looking variable
		# Returns:
		[value] or None
		"""
		try:
			cond = name in kwargs
			if(cond and "ignore_none" in kwargs and kwargs["ignore_none"]):
				cond = cond and kwargs[name] is not None
			return kwargs[name] if cond else self.parse_schema_value(self.schema[name])

		except:
			return None

	def convert_json(self, json_obj):
		"""
		Function to convert json to an acceptable serializable json
		# Parameters:
		----------
		- json_obj: dict
		Dictionary containing object ids, datetimes or another non serializable value
		# Returns:
		Serializable dictionary
		"""
		obj = json_util.dumps(json_obj, json_options=json_util.STRICT_JSON_OPTIONS)
		obj = json.loads (obj)
		return obj

	@classmethod
	def exists(cls, query):
		retval = methods.exists(cls.schema_name, query)
		return retval

	@classmethod
	def count(cls, query: dict) -> int:
		"""
		Count a number of documents inside a collection by query
		# Parameters
		------------
		### query: dict
			Dictionary containing query
		# Returns
		------------
		- int
		"""
		if methods.debug_log:
			Logger.printLog(cls.schema_name + " Count")
		retval = methods.count_documents(cls.schema_name, query)
		return retval

	@classmethod
	def find(cls, query, select = None, populate=None, one=False, skip = 0, limit=None, sort=None, parse=True):
		"""
		Find a document inside a collection
		# Parameters
		------------
		### query: dict
			Dictionary containing query
		### select: dict
			Dictionary containing requested fields
			defaults to: {}
		### populate: dict
			Dictionary containing lookup structure of the fields required to populate
			defaults to: None
		### one: bool
			Variable to select if return one ore many
			defaults to: False
		### skip: int
			Integer to skip to 'n' values to the left in the collection
			defaults to: 0
		### limit: int
			Integer to limit number of documents returned from the collection
			defaults to: None
		### sort: dict
			Dictionary to set an order of documents based on a field
			defaults to: None
		### parse: bool
			If one is True, then document could be parsed and returned as a Schema object
			defaults to: True
		# Returns
		------------
		- Cursor -> one == False, populate == None
		- CommandCursor -> populate != None
		- dict -> one == True, parse = False
		- Schema object -> one == True, parse == True
		- None -> not found
		"""
		if methods.debug_log:
			Logger.printLog(cls.schema_name)
		retval = methods.find(cls.schema_name, query, select, populate, one, skip, limit, sort)
		if one:
			retval = cls.parse(retval) if parse else retval
		return retval

	@classmethod
	def find_by_id(cls, id, select = None, populate=None, parse=True):
		"""
		Find a document inside a collection by id
		# Parameters
		------------
		### id: str | ObjectId
			Id of the document to lookup
		### select: dict
			Dictionary containing requested fields
			defaults to: {}
		### populate: dict
			Dictionary containing lookup structure of the fields required to populate
			defaults to: None
		### parse: bool
			If true, document will be returned as an Schema Object
			defaults to: True
		# Returns
		------------
		- dict -> parse = False
		- Schema object -> parse = True 
		"""
		retval = methods.find_by_id(cls.schema_name, id, select, populate)
		
		return cls.parse(retval) if parse else retval

	@classmethod
	def aggregate(cls, aggregate):
		"""
		Generate an Aggregate inside a collection
		# Parameters
		-------------
		### aggregate: list
		List of dicts following aggregation mongodb rules
		# Returns
		------------
		- commandCursor
		"""
		retval = methods.aggregate(cls.schema_name, aggregate)
		return retval

	@classmethod
	def update(cls, query, update, many = False):
		"""
		Update document(s) inside a collection by query
		# Parameters
		------------
		### query: dict
			Dictionary containing query
		### update: dict
			Dictionary containing update, following mongoDB rules
		### many: bool
			If True, will modify all documents found with that query
		defaults to False
		# Returns
		------------
		- int -> Modified count
		"""
		retval = methods.update(cls.schema_name, query, update, many)
		return retval

	@classmethod
	def delete(cls, query, many = False):
		"""
		Update document(s) inside a collection by query
		# Parameters
		------------
		### query: dict
			Dictionary containing query
		### many: bool
			If True, will modify all documents found with that query
		defaults to False
		# Returns
		------------
		- int -> Modified count
		"""
		retval = methods.delete(cls.schema_name, query, many)
		return retval

	@classmethod
	def parse(cls, dictionary):
		"""
		Converts a dictionary into a Schema object
		# Parameters
		------------
		### dictionary: dict
			Dictionary containing document information
		# Returns
		------------
		- Schema object -> Result
		"""
		schema = cls()
		schema.fromJson(dictionary)
		return schema


PYMONGOOSE_VERSION = "0.1"
PYMONGOOSE_VERSION_NAME = "Version 0.1"
PYMONGOOSE_VERSION_AUTHOR = "Juan Carlos Lara"

def pymongoose_version_print_full():
    print(f"{PYMONGOOSE_VERSION} - {PYMONGOOSE_VERSION_NAME}\nBy: {PYMONGOOSE_VERSION_AUTHOR}")